unit CosineTable;

interface

const
  GDefaultCosineTableSize=$FFFF;

type
  TFloat=Extended;

function TCos(X:TFloat):TFloat;
function TSin(X:TFloat):TFloat;
procedure TSinCos(X:TFloat;var S,C:TFloat);

procedure AllocateCosineTable(Size:Integer=0);

implementation

type
  TFloats=array[0..$FFFFFF] of TFloat;
  PFloats=^TFloats;

var
  GCosineTableSize:Integer=0;
  GCosineTableData:PFloats=nil;

procedure UnallocateCosineTable;
begin
  if GCosineTableSize>0 then begin
    FreeMem(GCosineTableData);
    GCosineTableData:=nil;
    GCosineTableSize:=0;
  end;
end;

procedure AllocateCosineTable(Size:Integer=0);
var
  a:Integer;
begin
  if (GCosineTableSize=0) or (Size>GCosineTableSize) then begin
    UnallocateCosineTable;
    if Size=0 then
      Size:=GDefaultCosineTableSize;
    GetMem(GCosineTableData,(Size+1)*Sizeof(TFloat));
    for a:=0 to Size do
      GCosineTableData[a]:=Cos(0.5*Pi*a/Size);
    GCosineTableSize:=Size;
  end;
end;

function TCos(X:TFloat):TFloat;
var
  N:Integer;
begin
  AllocateCosineTable;
  N:=Round(GCosineTableSize*4*Frac(Abs(X/(2*Pi))));
  if N>2*GCosineTableSize then
    N:=4*GCosineTableSize-N;
  if N>GCosineTableSize then
    Result:=-GCosineTableData[2*GCosineTableSize-N]
  else
    Result:=GCosineTableData[N];
end;

function TSin(X:TFloat):TFloat;
begin
  Result:=TCos(0.5*Pi-X);
end;

procedure TSinCos(X:TFloat;var S,C:TFloat);
var
  N:Integer;
  u:Integer;
begin
  AllocateCosineTable;
  if X>0 then begin
    u:=1;
    N:=Round(GCosineTableSize*4*Frac(X/(2*Pi)));
  end else begin
    u:=-1;
    N:=Round(GCosineTableSize*4*Frac(-X/(2*Pi)));
  end;
  if N>2*GCosineTableSize then begin
    u:=-u;
    N:=4*GCosineTableSize-N;
  end;
  if N>GCosineTableSize then begin
    C:=-GCosineTableData[2*GCosineTableSize-N];
    S:=u*GCosineTableData[N-GCosineTableSize];
  end else begin
    C:=GCosineTableData[N];
    S:=u*GCosineTableData[GCosineTableSize-N];
  end;
end;

initialization
finalization
  UnallocateCosineTable;
end.

